local Vector3 = require "types.vector3"

---@class AABB
local AABB = {
  min = Vector3(),
  size = Vector3(),
}

local RED = {
  r = 255,
  g = 0,
  b = 0,
  a = 255,
}

setmetatable(AABB, AABB)

AABB.__index = AABB

---@param position Vector3
---@param size Vector3
---@return AABB
function AABB.new(position, size)
  position = position or Vector3()
  size = size or Vector3(1, 1, 1)
  local aabb = {
    min = position,
    size = size,
  }

  return setmetatable(aabb, AABB)
end

function AABB:get_max()
  return self.min + self.size
end

function AABB:draw()
  local max = self:get_max()
  -- bottom rectangle
  draw_line_3d{start = self.min, finish = Vector3(max.x, self.min.y, self.min.z)}
  draw_line_3d{start = self.min, finish = Vector3(self.min.x, self.min.y, max.z)}
  draw_line_3d{start = Vector3(max.x, self.min.y, max.z), finish = Vector3(self.min.x, self.min.y, max.z)}
  draw_line_3d{start = Vector3(max.x, self.min.y, max.z), finish = Vector3(max.x, self.min.y, self.min.z)}
  -- bottom rectangle diagonal
  draw_line_3d{start = self.min, finish = Vector3(max.x, self.min.y, max.z), color = RED}
  -- top rectangle
  draw_line_3d{start = Vector3(self.min.x, max.y, self.min.z), finish = Vector3(max.x, max.y, self.min.z)}
  draw_line_3d{start = Vector3(self.min.x, max.y, self.min.z), finish = Vector3(self.min.x, max.y, max.z)}
  draw_line_3d{start = Vector3(max.x, max.y, max.z), finish = Vector3(self.min.x, max.y, max.z)}
  draw_line_3d{start = Vector3(max.x, max.y, max.z), finish = Vector3(max.x, max.y, self.min.z)}
  -- top rectangle diagonal
  draw_line_3d{start = Vector3(max.x, max.y, max.z), finish = Vector3(self.min.x, max.y, self.min.z), color = RED}
  -- hull
  draw_line_3d{start = self.min, finish = Vector3(self.min.x, max.y, self.min.z)}
  draw_line_3d{start = self.min, finish = Vector3(max.x, max.y, self.min.z), color = RED}

  draw_line_3d{start = Vector3(max.x, self.min.y, self.min.z), finish = Vector3(max.x, max.y, self.min.z)}
  draw_line_3d{start = Vector3(max.x, self.min.y, self.min.z), finish = Vector3(max.x, max.y, max.z), color = RED}

  draw_line_3d{start = Vector3(max.x, self.min.y, max.z), finish = Vector3(max.x, max.y, max.z)}
  draw_line_3d{start = Vector3(max.x, self.min.y, max.z), finish = Vector3(self.min.x, max.y, max.z), color = RED}

  draw_line_3d{start = Vector3(self.min.x, self.min.y, max.z), finish = Vector3(self.min.x, max.y, max.z)}
  draw_line_3d{start = Vector3(self.min.x, self.min.y, max.z), finish = Vector3(self.min.x, max.y, self.min.z), color = RED}
end

---returns true if the point is inside this aabb
---@param point Vector3
---@return boolean
function AABB:has_point(point)
  local max = self:get_max()
  return point > self.min and point < max
end

---returns true if `other` intersects this AABB
---@param other AABB
---@return boolean
function AABB:intersects(other)
  local my_max = self:get_max()
  local other_max = other:get_max()
  return self.min <= other_max and my_max >= other.min
end

return AABB
