---@class Signal
---@field private _connections table
local Signal = {}

Signal.__index = Signal

---Constructs a new signal.
---@return Signal
function Signal.new()
  local s = {
    _connections = {},
  }
  
  return setmetatable(s, Signal)
end

---Connects f to this signal. When the signal is emmitted, this function will be called.
---@param f function
function Signal:connect(f)
  util.list_insert_once(self._connections, f)
end

---Disconnects f from this signal.
---@param f function
function Signal:disconnect(f)
  util.list_remove_value(self._connections, f)
end

---Emits the signal, calling the connecting functions with the provided params.
---@param ... any
function Signal:emit(...)
  -- don't care about order
  for _, v in pairs(self._connections) do
    v(...)
  end
end

return Signal